#![no_std]
#![no_main]
extern crate alloc;

use blocking_network_stack::Stack;

use embedded_io::*;

use esp_backtrace as _;
use esp_hal::{
    clock::CpuClock, delay::Delay, rng::Rng, time::{self, Duration}, timer::timg::TimerGroup, xtensa_lx_rt::entry
};


use esp_wifi::wifi::{utils::create_network_interface, AccessPointInfo, AuthMethod, ClientConfiguration, Configuration, WifiError, WifiStaDevice};
use smoltcp::iface::{SocketSet, SocketStorage};

fn parse_ip(ip: &str) -> [u8; 4] {
    let mut result = [0u8; 4];
    for (idx, octet) in ip.split(".").into_iter().enumerate() {
        result[idx] = u8::from_str_radix(octet, 10).unwrap();
    }
    result
}

#[entry]
fn main() -> ! {

    // Initialize the ESP-IDF logging system
    esp_println::logger::init_logger(log::LevelFilter::Info);

    // Setup the cpu clock config and initialize
    let config = esp_hal::Config::default().with_cpu_clock(CpuClock::max());
    let peripherals = esp_hal::init(config);

    // Create a delay object
    let delay = Delay::new();

    // You need to allocate here and you must use the same version as esp-wifi 
    esp_alloc::heap_allocator!(72 * 1024);

    // Initialize the timer group
    let timg0 = TimerGroup::new(peripherals.TIMG0);

    // Initialize the random number generator   
    let mut rng = Rng::new(peripherals.RNG);

    // Initialize the wifi controller
    let esp_wifi_ctrl = esp_wifi::init(
        timg0.timer0,
        rng.clone(),
        peripherals.RADIO_CLK,
    )
    .unwrap();

    // Set credentials from environment variables
    const SSID: &str = env!("SSID");
    const PASSWORD: &str = env!("PASSWORD");
    const GATEWAY_IP: &str = env!("GATEWAY_IP");
    const STATIC_IP: &str = env!("STATIC_IP");

    // Get the wifi Peripheral
    let mut wifi = peripherals.WIFI;

    // Initialize the iface and device and the controller   
    let (iface, device, mut controller) =
         create_network_interface(&esp_wifi_ctrl, &mut wifi, WifiStaDevice).unwrap();
    
    // Create the socket set which is like the c struct sockaddr
    let mut socket_set_entries: [SocketStorage; 3] = Default::default();
    let socket_set = SocketSet::new(&mut socket_set_entries[..]);

    // Create the stack
    let now = || time::now().duration_since_epoch().to_millis();
    let mut wifi_stack = Stack::new(iface, device, socket_set, now, rng.random());

    // Configure Authentication 
    let mut auth_method = AuthMethod::WPA2Personal;
    if PASSWORD.is_empty() {
         auth_method = AuthMethod::None;
    }
    
    // Configure the Wi-Fi client       
    let client_config = Configuration::Client(ClientConfiguration {
         ssid: SSID.try_into().unwrap(),
         password: PASSWORD.try_into().unwrap(),
         auth_method,
         ..Default::default()
     });

    let res = controller.set_configuration(&client_config);
    log::info!("Wi-Fi set_configuration returned {:?}", res);

    controller.start().unwrap();
    log::info!("Is wifi started: {:?}", controller.is_started());

    log::info!("Starting Wifi Scan");   

    // Scan for APs
    let res: Result<(heapless::Vec<AccessPointInfo, 10>, usize), WifiError> = controller.scan_n();
    if let Ok((res, _count)) = res {
        for ap in res {
            log::info!("{:?}", ap);
        }
    }    

    log::info!("{:?}", controller.capabilities());
    log::info!("Wi-Fi connect: {:?}", controller.connect());

    // wait to get connected
    log::info!("Waiting to get connected");   
    loop {
        let res = controller.is_connected();
        match res {
            Ok(connected) => {
                if connected {
                    break;
                }
            }
            Err(err) => {
                log::info!("{:?}", err);
                loop {}
            }
        }
    }
    log::info!("{:?}", controller.is_connected());

    log::info!("Setting static IP {}", STATIC_IP);

    wifi_stack
        .set_iface_configuration(&blocking_network_stack::ipv4::Configuration::Client(
            blocking_network_stack::ipv4::ClientConfiguration::Fixed(
                blocking_network_stack::ipv4::ClientSettings {
                    ip: blocking_network_stack::ipv4::Ipv4Addr::from(parse_ip(STATIC_IP)),
                    subnet: blocking_network_stack::ipv4::Subnet {
                        gateway: blocking_network_stack::ipv4::Ipv4Addr::from(parse_ip(GATEWAY_IP)),
                        mask: blocking_network_stack::ipv4::Mask(24),
                    },
                    dns: None,
                    secondary_dns: None,
                },
            ),
        ))
        .unwrap();    

    log::info!(
        "Start busy loop on main. Point your browser to http://{}:8080/",
        STATIC_IP
    );

    let mut rx_buffer = [0u8; 1536];
    let mut tx_buffer = [0u8; 1536];
    let mut socket = wifi_stack.get_socket(&mut rx_buffer, &mut tx_buffer);

    socket.listen(8080).unwrap();

    loop {
        socket.work();

        if !socket.is_open() {
            socket.listen(8080).unwrap();
        }

        if socket.is_connected() {
            log::info!("Connected");

            let mut time_out = false;
            let deadline = time::now() + Duration::secs(20);
            let mut buffer = [0u8; 1024];
            let mut pos = 0;

            while let Ok(len) = socket.read(&mut buffer[pos..]) {
                let to_print = unsafe { core::str::from_utf8_unchecked(&buffer[..(pos + len)]) };

                if to_print.contains("\r\n\r\n") {
                    log::info!("{}", to_print);
                    break;
                }

                pos += len;

                if time::now() > deadline {
                    log::info!("Timeout");
                    time_out = true;
                    break;
                }
            }

            if !time_out {
                socket.write_all(
                    b"HTTP/1.0 200 OK\r\n\r\n\
                    <html>\
                        <body>\
                            <h1>Hello Rust! Hello esp-wifi!</h1>\
                            <img src=\"https://rustacean.net/more-crabby-things/dancing-ferris.gif\"/>
                        </body>\
                    </html>\r\n\
                    "
                ).unwrap();

                socket.flush().unwrap();
            }

            socket.close();

            log::info!("Done");
        }

        let deadline = time::now() + Duration::secs(5);
        while time::now() < deadline {
            socket.work();
        }
    }    
}
